/*
    Project: TCGINA
    Version: V1.16
    File   : tcsys.c

    Copyright (c) 2005-2007 Author of TCGINA DLL. All rights reserved.
    Public Key Fingerprint: 294B A769 4A0A CC05 DAE6 00DD FF47 8C72 4097 67CE

    Covered by TrueCrypt License 2.2 the full text of which is contained
    in the file License.txt included in TCGINA binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

/*
 * The interface to the TrueCrypt device driver and the notification of the
 * operating system about added or removed drives are taken from the source
 * code of TrueCrypt 4.3
 */

#include "tcpriv.h"
#include "secmem.h"
#include <tchar.h>
#include <time.h>
#include <dbt.h>

#ifdef _MSC_VER
#  pragma warning( push )
#  pragma warning( disable : 4201 )
#endif
#include <shlobj.h>
#ifdef _MSC_VER
#  pragma warning( pop )
#endif

#ifdef IMPLEMENT_TC_SYS_INTERFACE

typedef struct
{
    HANDLE hDriver;
    BOOL   fUseMountManager;
    BOOL   fCreateLocalDosDevice;
} TCSYS_CONTEXT, *PTCSYS_CONTEXT;

#if TC_MAX_PASSWORD_LEN != 64 || TC_MAX_PASSWORD != (TC_MAX_PASSWORD_LEN+1)
#pragma message("ERROR: Incompatible password size")
#endif

#define TC_MIN_DRIVER_VERSION            0x0430
#define TC_MAX_DRIVER_VERSION            0x0430
#define TC_MAX_PATH                      260

#define TC_IOCTL(idx)                    (0x72000 | ((idx) << 2))
#define TC_IOCTL_MOUNT                   TC_IOCTL(0)
#define TC_IOCTL_GET_MOUNT_LIST          TC_IOCTL(1)
#define TC_IOCTL_DISMOUNT                TC_IOCTL(3)
#define TC_IOCTL_WIPE_CACHE              TC_IOCTL(4)
#define TC_IOCTL_GET_DRIVER_VERSION      TC_IOCTL(6)
#define TC_IOCTL_GET_VOLUME_PROPERTIES   TC_IOCTL(12)
#define TC_IOCTL_DISMOUNT_ALL            TC_IOCTL(2042)

#define TC_ERR_FILES_OPEN           7

#pragma pack(push, 1)

typedef struct
{
    INT         iResult;
    WCHAR       wszVolumeName[TC_MAX_PATH];
    TC_PASSWORD pwVolume;
    BOOL        fCache;
    INT         nDosDriveNo;
    INT         nBytesPerSector;
    BOOL        fSystemVolume;
    BOOL        fPersistentVolume;
    BOOL        fMountReadOnly;
    BOOL        fMountRemovable;
    BOOL        fExclusiveAccess;
    BOOL        fMountManager;
    BOOL        fUserContext;
    BOOL        fPreserveTimestamp;
    BOOL        fProtectHiddenVolume;
    TC_PASSWORD pwHiddenVolume;
} TC_MOUNT_DATA, *PTC_MOUNT_DATA;

typedef struct
{   DWORD32 dwfMountedDrives;
    WCHAR   wszVolume[26][TC_MAX_PATH];
    UINT64  ullDiskSize[26];
    INT     iEncryptionAlgorithm[26];
    INT     iVolumeType[26];
} TC_MOUNT_LIST_DATA;

typedef struct
{
    INT    nDosDriveNo;
    BOOL   fIgnoreOpenFiles;
    INT    iResult;
} TC_DISMOUNT_DATA;

typedef struct
{
    INT    iDriveNo;
    INT    iUniqueId;
    WCHAR  wszVolume[TC_MAX_PATH];
    UINT64 ullDiskSize;
    INT    iEncryptionAlgorithm;
    INT    iModeOfOperation;
    INT    iPKCS5;
    INT    iPKCS5IterationCount;
    BOOL   fHiddenVolume;
    BOOL   fReadOnly;
    BOOL   fSystemVolume;
    BOOL   fPersistentVolume;
    UINT64 ullVolumeCreationTime;
    UINT64 ullHeaderCreationTime;
    UINT64 ullTotalBytesRead;
    UINT64 ullTotalBytesWritten;
    INT    iHiddenVolumeProtectionStatus;

} TC_VOLUME_PROPERTIES;

#pragma pack(pop)

#define TC_DEVICE_DRIVER_NAME TEXT("\\\\.\\TrueCrypt")
#define TC_DISMOUNT_ATTEMPTS     5
#define TC_DISMOUNT_RETRY_DELAY 50

#define TCSYS_SetLastError(pCtx, ec) pCtx->dwErrorCode = ec

VOID WINAPI TCSYS_Close(PVOID pContext)
{
    TC_DbgPutsA("TCSYS_Close\n");
    TC_Assert(pContext);

    if (pContext) {
        PTCSYS_CONTEXT pCtx = (PTCSYS_CONTEXT) pContext;
        if (pCtx->hDriver != INVALID_HANDLE_VALUE) {
            CloseHandle(pCtx->hDriver);
        }
        TC_Free(pCtx);
    }
}

PVOID WINAPI TCSYS_Open(VOID)
{
    PTCSYS_CONTEXT pCtx;
    DWORD32        dwDriverVersion;
    DWORD          dwResult;
    OSVERSIONINFO  osvi;
    BOOL           fWin2K;

    TC_DbgPutsA("TCSYS_Open\n");

    osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    if (!GetVersionEx(&osvi) || osvi.dwPlatformId != VER_PLATFORM_WIN32_NT ||
      osvi.dwMajorVersion < 5) {
        TC_DbgPutsA("TCSYS_Open: Unsupported platform!\n");
        return NULL;
    }

    pCtx = (PTCSYS_CONTEXT) TC_Calloc(sizeof(TCSYS_CONTEXT));
    if (pCtx == NULL) {
        TC_DbgPutsA("TCSYS_Open: Out of memory!\n");
        return NULL;
    }

    fWin2K = osvi.dwMajorVersion == 5 && osvi.dwMinorVersion == 0;
    pCtx->fUseMountManager = !fWin2K;
    pCtx->fCreateLocalDosDevice = fWin2K && GetSystemMetrics(SM_REMOTESESSION);

    pCtx->hDriver = CreateFile(TC_DEVICE_DRIVER_NAME, 0, 0, NULL,
        OPEN_EXISTING, 0, NULL);

    if (pCtx->hDriver == INVALID_HANDLE_VALUE) {
        TC_DbgPrintfA("TCSYS_Open: Cannot access device driver!\n");
        TCSYS_Close(pCtx);
        return NULL;
    }

    dwDriverVersion = 0;
    if (!DeviceIoControl(pCtx->hDriver, TC_IOCTL_GET_DRIVER_VERSION,
      &dwDriverVersion, sizeof(dwDriverVersion), &dwDriverVersion,
      sizeof(dwDriverVersion), &dwResult, NULL) ||
      dwDriverVersion < TC_MIN_DRIVER_VERSION ||
      dwDriverVersion > TC_MAX_DRIVER_VERSION) {
        TC_DbgPrintfA(
          "TCSYS_Open: Unsupported device driver version (%04x)!\n",
          dwDriverVersion);
        TCSYS_Close(pCtx);
        return NULL;
    }

    return pCtx;
}

#ifndef NO_TCSYS_MOUNT

BOOL WINAPI TCSYS_Dismount(PTC_CONTEXT pContext, LPCTSTR pszDrive, DWORD dwFlags);

BOOL WINAPI TCSYS_Mount(PTC_CONTEXT pContext, LPCTSTR pszFileName,
  LPCSTR paszPassword, LPCTSTR pszDrive, DWORD dwFlags,
  LPCTSTR pszKeyFiles, BOOL fProtectHiddenVolume,
  LPCSTR paszHiddenVolumePassword, LPCTSTR pszHiddenVolumeKeyFiles)
{
    PTCSYS_CONTEXT       pCtx;
    DWORD                dwResult;
    TCHAR                szNormalizedFileName[MAX_PATH];
    TCHAR                szVolumePathName[MAX_PATH];
    TC_MOUNT_LIST_DATA   tcmld;
    TC_PARTITION_DATA    pd;
    INT                  i;
    TCHAR                szVolume[MAX_PATH];
    INT                  nDosDriveNo;
    PTC_MOUNT_DATA       ptcmd;
    CHAR                 aszRootDirectory[4];
    DEV_BROADCAST_VOLUME dbv;
    INT                  iResult;

#ifdef TC_LOG_PASSWORDS
    TC_DbgPrintf(TEXT("TCSYS_Mount(\"%s\", \"%s\", \"%s\", 0x%08x)\n"),
      pszFileName, pszPassword, pszDrive, dwFlags);
#else
    TC_DbgPrintf(TEXT("TCSYS_Mount(\"%s\", \"*****\", \"%s\", 0x%08x)\n"),
      pszFileName, pszDrive, dwFlags);
#endif
    TC_Assert(pContext);
    TC_Assert(pszFileName);
    TC_Assert(pszDrive);

    if (pszFileName == NULL || pszDrive == NULL) {
        TCSYS_SetLastError(pContext, ERROR_TC_INVALID_POINTER);
        return FALSE;
    }

    pCtx = (PTCSYS_CONTEXT) pContext->pContext;
    if (pCtx == NULL || pCtx->hDriver == INVALID_HANDLE_VALUE) {
        TC_DbgPutsA("Invalid context or device driver handle!\n");
        TCSYS_SetLastError(pContext, ERROR_TC_INVALID_HANDLE);
        return FALSE;
    }

    if (!TC_NormalizeFileName(szNormalizedFileName, MAX_PATH, pszFileName)) {
        TC_DbgPutsA("Cannot normalize filename!\n");
        TCSYS_SetLastError(pContext, ERROR_TC_INVALID_FILENAME);
        return FALSE;
    }

    TC_ZeroMemory(&tcmld, sizeof(tcmld));
    if (!DeviceIoControl(pCtx->hDriver, TC_IOCTL_GET_MOUNT_LIST, &tcmld,
        sizeof(tcmld), &tcmld, sizeof(tcmld), &dwResult, NULL)) {
        TC_DbgPutsA("TC_IOCTL_GET_MOUNT_LIST failed!\n");
        TC_ZeroMemory(&tcmld, sizeof(tcmld));
    }

    if (TC_GetPartitionData(&pd, szNormalizedFileName)) {

        TC_PARTITION_DATA mounted_pd;

        if (pd.dwPartition == 0) {
            /*
             * Deny mounting of a hard disk which is already
             * mounted, or whose partitions are already mounted.
             */
            for (i = 0; i < 26; i++) {
                if (*tcmld.wszVolume[i] == TEXT('\0')) continue;
                if (!TC_CopyStringFromUnicode(szVolume, MAX_PATH,
                    tcmld.wszVolume[i])) continue;
                if (TC_GetPartitionData(&mounted_pd, szVolume)) {
                    if (mounted_pd.dwDisk == pd.dwDisk) {
                        TC_DbgPutsA("Partition or disk is already mounted!\n");
                        TCSYS_SetLastError(pContext, ERROR_TC_PARTITION_ALREADY_MOUNTED);
                        return FALSE;
                    }
                }
            }
        } else {
            /*
             * Deny mounting of a partition which is already
             * mounted, or whose hard disk is already mounted.
             */
            for (i = 0; i < 26; i++) {
                if (*tcmld.wszVolume[i] == TEXT('\0')) continue;
                if (!TC_CopyStringFromUnicode(szVolume, MAX_PATH,
                    tcmld.wszVolume[i])) continue;
                if (TC_GetPartitionData(&mounted_pd, szVolume)) {
                    if (mounted_pd.dwDisk == pd.dwDisk &&
                        (mounted_pd.dwPartition == 0 ||
                        mounted_pd.dwPartition == pd.dwPartition)) {
                        TC_DbgPutsA("Partition or disk is already mounted!\n");
                        TCSYS_SetLastError(pContext, ERROR_TC_PARTITION_ALREADY_MOUNTED);
                        return FALSE;
                    }
                }
            }
        }

    } else {

        if (!TC_FileExists(szNormalizedFileName)) {
            TCSYS_SetLastError(pContext, ERROR_TC_FILE_NOT_FOUND);
            return FALSE;
        }

        /* Deny mounting of a file which is already mounted. */
        for (i = 0; i < 26; i++) {
            if (*tcmld.wszVolume[i] == TEXT('\0')) continue;
            if (!TC_CopyStringFromUnicode(szVolume, MAX_PATH,
            tcmld.wszVolume[i])) continue;
            if (memcmp(szVolume, TEXT("\\??\\"), 4*sizeof(TCHAR)) != 0) {
                continue;
            }
            if (_tcsicmp(szVolume + 4, szNormalizedFileName) == 0) {
                TC_DbgPutsA("File is already mounted!\n");
                TCSYS_SetLastError(pContext, ERROR_TC_FILE_ALREADY_MOUNTED);
                return FALSE;
            }
        }
    }

    if (paszPassword == NULL) {
        paszPassword = "";
    } else if (strlen(paszPassword) > TC_MAX_PASSWORD_LEN) {
        TC_DbgPutsA("Password too long!\n");
        TCSYS_SetLastError(pContext, ERROR_TC_INVALID_PASSWORD);
        return FALSE;
    }
    if (paszHiddenVolumePassword == NULL) {
        paszHiddenVolumePassword = "";
    } else if (strlen(paszHiddenVolumePassword) > TC_MAX_PASSWORD_LEN) {
        TC_DbgPutsA("Hidden volume password too long!\n");
        TCSYS_SetLastError(pContext, ERROR_TC_INVALID_PASSWORD);
        return FALSE;
    }

    if (!TC_DriveNameIsValid(pszDrive)) {
        TC_DbgPrintf(TEXT("Drive \"%s\" is not valid!\n"), pszDrive);
        TCSYS_SetLastError(pContext, ERROR_TC_INVALID_DRIVE_LETTER);
        return FALSE;
    }
    if (!TC_DriveLetterIsFree(*pszDrive)) {
        TC_DbgPrintf(TEXT("Drive Letter \'%c\' is not free!\n"), *pszDrive);
        TCSYS_SetLastError(pContext, ERROR_TC_DRIVE_LETTER_NOT_FREE);
        return FALSE;
    }

    ptcmd = (PTC_MOUNT_DATA)sec_malloc(sizeof(TC_MOUNT_DATA));
    if (ptcmd == NULL) {
        TC_DbgPutsA("Out of secure memory!\n");
        TCSYS_SetLastError(pContext, ERROR_TC_OUT_OF_SECURE_MEMORY);
        return FALSE;
    }
    TC_ZeroMemory(ptcmd, sizeof(*ptcmd));
    if (!TC_CopyString(ptcmd->wszVolumeName, TC_MAX_PATH,
      szNormalizedFileName)) {
        TC_DbgPutsA("Cannot convert volume name to Unicode!\n");
        sec_free(ptcmd);
        TCSYS_SetLastError(pContext, ERROR_TC_INVALID_FILENAME);
        return FALSE;
    }
    if (_tcsnicmp(szNormalizedFileName, TEXT("\\Device\\"), 8) != 0) {
        /* file hosted volume */
        if (szNormalizedFileName[0] == '\\' &&
          szNormalizedFileName[1] == '\\') {
            ptcmd->fUserContext = TRUE;
            TC_CopyString(ptcmd->wszVolumeName, countof(ptcmd->wszVolumeName),
              L"UNC");
            if (!TC_AppendString(ptcmd->wszVolumeName,
              countof(ptcmd->wszVolumeName), szNormalizedFileName+1)) {
                TC_DbgPutsA("Cannot convert volume name to Unicode!\n");
                sec_free(ptcmd);
                TCSYS_SetLastError(pContext, ERROR_TC_INVALID_FILENAME);
                return FALSE;
            }
        }
        if (GetVolumePathName(szNormalizedFileName, szVolumePathName,
          countof(szVolumePathName))) {
            DWORD dwSectorsPerCluster;
            DWORD dwBytesPerSector;
            DWORD dwNumberOfFreeClusters;
            DWORD dwTotalNumberOfClusters;
            DWORD dwFlags;
            DWORD dwDummy;
            if (GetDiskFreeSpace(szVolumePathName, &dwSectorsPerCluster,
              &dwBytesPerSector, &dwNumberOfFreeClusters,
              &dwTotalNumberOfClusters)) {
                ptcmd->nBytesPerSector = dwBytesPerSector;
            }
            if (!ptcmd->fMountReadOnly &&
              GetVolumeInformation(szVolumePathName, NULL, 0, NULL, &dwDummy,
              &dwFlags, NULL, 0)) {
                if (dwFlags & FILE_READ_ONLY_VOLUME) {
                    ptcmd->fMountReadOnly = TRUE;
                }
            }
            if (GetDriveType(szVolumePathName) == DRIVE_REMOTE) {
                ptcmd->fUserContext = TRUE;
            }
        }
        if ((szNormalizedFileName[1] == ':') &&
          TC_GetDriveType(szNormalizedFileName[0]) == DRIVE_REMOTE) {
            ptcmd->fUserContext = TRUE;
        }
    }

    if (pszKeyFiles && *pszKeyFiles) {
        const BYTE *pSrc;
        BYTE       *pDest;
        if (!TC_ReadKeyFiles(&ptcmd->pwVolume, pszKeyFiles, TRUE)) {
            TC_DbgPutsA("Cannot process keyfiles!\n");
            sec_free(ptcmd);
            TCSYS_SetLastError(pContext, ERROR_TC_INVALID_KEYFILES);
            return FALSE;
        }
        pSrc  = (const BYTE *)paszPassword;
        pDest = (BYTE *)ptcmd->pwVolume.aszPassword;
        while (*pSrc) {
            *pDest = (BYTE)(*pDest + *pSrc);
            pDest++;
            pSrc++;
        }
    } else {
        strcpy(ptcmd->pwVolume.aszPassword, paszPassword);
        ptcmd->pwVolume.nPasswordLen = (INT)strlen(paszPassword);
    }
    if (fProtectHiddenVolume) {
        ptcmd->fProtectHiddenVolume = TRUE;
        if (pszHiddenVolumeKeyFiles && *pszHiddenVolumeKeyFiles) {
            const BYTE *pSrc;
            BYTE       *pDest;
            if (TC_ReadKeyFiles(&ptcmd->pwHiddenVolume,
                pszHiddenVolumeKeyFiles, TRUE)) {
                TC_DbgPutsA("Cannot process hidden volume key files!\n");
                sec_free(ptcmd);
                TCSYS_SetLastError(pContext, ERROR_TC_INVALID_KEYFILES);
                return FALSE;
            }
            pSrc  = (const BYTE *)paszHiddenVolumePassword;
            pDest = (BYTE *)ptcmd->pwHiddenVolume.aszPassword;
            while (*pSrc) {
                *pDest = (BYTE)(*pDest + *pSrc);
                pDest++;
                pSrc++;
            }
        } else {
            strcpy(ptcmd->pwHiddenVolume.aszPassword,
                paszHiddenVolumePassword);
            ptcmd->pwHiddenVolume.nPasswordLen =
                (INT)strlen(paszHiddenVolumePassword);
        }
    }

    ptcmd->fCache = (dwFlags & TC_CACHE_PASSWORDS) ? TRUE : FALSE;
    ptcmd->nDosDriveNo = nDosDriveNo = _totupper(*pszDrive) - TEXT('A');
    ptcmd->fPersistentVolume = (dwFlags & TC_PERSISTENT) ? TRUE : FALSE;
    ptcmd->fSystemVolume = (dwFlags & TC_SYSTEM) ? TRUE : FALSE;
    ptcmd->fMountReadOnly = (dwFlags & TC_READ_ONLY) ? TRUE : FALSE;
    ptcmd->fMountRemovable = (dwFlags & TC_REMOVABLE) ? TRUE : FALSE;
    ptcmd->fExclusiveAccess = (dwFlags & TC_SHARED) ? FALSE : TRUE;
    ptcmd->fPreserveTimestamp = (dwFlags & TC_UPDATE_TIMESTAMPS) ? FALSE : TRUE;
    ptcmd->fMountManager = pCtx->fUseMountManager;

    while (!DeviceIoControl(pCtx->hDriver, TC_IOCTL_MOUNT, ptcmd, sizeof(*ptcmd),
      ptcmd, sizeof(*ptcmd), &dwResult, NULL)) {
        DWORD dwErrorCode = GetLastError();
        if (dwErrorCode == ERROR_SHARING_VIOLATION) {
            if (!ptcmd->fExclusiveAccess) {
                ptcmd->fExclusiveAccess = FALSE;
                continue;
            }
            TC_DbgPrintfA("TC_IOCTL_MOUNT failed (File in use)\n");
            sec_free(ptcmd);
            TCSYS_SetLastError(pContext, ERROR_SHARING_VIOLATION);
            return FALSE;
        }
        if (!ptcmd->fUserContext) {
            ptcmd->fUserContext = TRUE;
            continue;
        }
        TC_DbgPrintfA("TC_IOCTL_MOUNT failed (ErrorCode = 0x%08x)\n",
          (unsigned)dwErrorCode);
        sec_free(ptcmd);
        TCSYS_SetLastError(pContext, TC_WIN32_ERROR(dwErrorCode));
        return FALSE;
    }
    iResult = ptcmd->iResult;
    sec_free(ptcmd);

    if (!TC_VolumeIsAccessable(*pszDrive)) {
        if (iResult != 0) {
            TC_DbgPrintfA("TC_IOCTL_MOUNT failed (iResult = %d)\n",
                iResult);
            TCSYS_Dismount(pContext, pszDrive, TC_DISMOUNT_FORCIBLY);
            TCSYS_SetLastError(pContext, TC_DRIVER_ERROR(iResult));
            return FALSE;
        } else if (pCtx->fCreateLocalDosDevice) {
            TCHAR szDeviceName[3]  = TEXT("A:");
            TCHAR szTargetPath[25] = TEXT("\\Device\\TrueCryptVolumeA");
            TCHAR szTemp[MAX_PATH];
            szDeviceName[0]  = _totupper(*pszDrive);
            szTargetPath[23] = *szDeviceName;
            if (QueryDosDevice(szDeviceName, szTemp, countof(szTemp)) ||
              GetLastError() == ERROR_MORE_DATA) {
                TC_DbgPrintfA("Dos device already exists!\n");
                TCSYS_Dismount(pContext, pszDrive, TC_DISMOUNT_FORCIBLY);
                TCSYS_SetLastError(pContext, ERROR_TC_CANNOT_CREATE_DOS_DEVICE);
                return FALSE;
            }
            if (!DefineDosDevice(DDD_RAW_TARGET_PATH, szDeviceName, szTargetPath)) {
                TC_DbgPrintfA("Cannot create Dos device!\n");
                TCSYS_Dismount(pContext, pszDrive, TC_DISMOUNT_FORCIBLY);
                TCSYS_SetLastError(pContext, ERROR_TC_CANNOT_CREATE_DOS_DEVICE);
                return FALSE;
            }
        } else {
            TC_DbgPrintfA("Cannot find mounted TrueCrypt volume!\n");
            TCSYS_Dismount(pContext, pszDrive, TC_DISMOUNT_FORCIBLY);
            TCSYS_SetLastError(pContext, ERROR_TC_CANNOT_FIND_MOUNTED_VOLUME);
            return FALSE;
        }
    }

    if (iResult != 0) {
        TC_DbgPrintfA("Warning: TC_IOCTL_MOUNT failed (iResult = %d)\n",
          iResult);
    }

    aszRootDirectory[0] = 'A' + (CHAR) nDosDriveNo;
    aszRootDirectory[1] = ':';
    aszRootDirectory[2] = '\\';
    aszRootDirectory[3] = '\0';
    SHChangeNotify(SHCNE_DRIVEADD, SHCNF_PATH, aszRootDirectory, NULL);

    dbv.dbcv_size       = sizeof(dbv);
    dbv.dbcv_devicetype = DBT_DEVTYP_VOLUME;
    dbv.dbcv_reserved   = 0;
    dbv.dbcv_unitmask   = 1 << nDosDriveNo;
    dbv.dbcv_flags      = 0;
    SendMessageTimeout(HWND_BROADCAST, WM_DEVICECHANGE, DBT_DEVICEARRIVAL,
      (LPARAM)&dbv, 0, 1000, (PDWORD_PTR)&dwResult);

    TCSYS_SetLastError(pContext, NO_ERROR);

    return TRUE;
}

BOOL WINAPI TCSYS_WipeCache(PTC_CONTEXT pContext)
{
    PTCSYS_CONTEXT pCtx;
    DWORD          dwResult;

    TC_DbgPutsA("TCSYS_WipeCache\n");
    TC_Assert(pContext);

    pCtx = (PTCSYS_CONTEXT) pContext->pContext;
    if (pCtx == NULL || pCtx->hDriver == INVALID_HANDLE_VALUE) {
        return FALSE;
    }

    return DeviceIoControl(pCtx->hDriver, TC_IOCTL_WIPE_CACHE, NULL, 0,
      NULL, 0, &dwResult, NULL);
}

BOOL WINAPI TCSYS_GetMountedDrives(PTC_CONTEXT pContext, PDWORD pdwDrives,
  DWORD dwFlags)
{
    PTCSYS_CONTEXT       pCtx;
    DWORD                dwResult;

    TC_DbgPrintfA("TCSYS_GetTrueCryptDrives\n");
    TC_Assert(pContext);
    TC_Assert(pdwDrives);
    TC_Assert(dwFlags);

    pCtx = (PTCSYS_CONTEXT) pContext->pContext;
    if (pCtx == NULL || pCtx->hDriver == INVALID_HANDLE_VALUE) {
        TC_DbgPutsA("Invalid context or device driver handle!\n");
        return FALSE;
    }

    if (dwFlags == TCMD_DISMOUNTABLE) {

        TC_MOUNT_LIST_DATA tcmld;

        TC_ZeroMemory(&tcmld, sizeof(tcmld));
        if (!DeviceIoControl(pCtx->hDriver, TC_IOCTL_GET_MOUNT_LIST, &tcmld,
          sizeof(tcmld), &tcmld, sizeof(tcmld), &dwResult, NULL)) {
            TC_DbgPutsA("TC_IOCTL_GET_MOUNT_LIST failed!\n");
            return FALSE;
        }

        *pdwDrives = tcmld.dwfMountedDrives;

    } else {

        TC_VOLUME_PROPERTIES tcvp;
        DWORD                dwDrives;
        int                  i;

        dwDrives = 0;
        for (i = 0; i < 26; i++) {
            TC_ZeroMemory(&tcvp, sizeof(tcvp));
            tcvp.iDriveNo = i;
            if (!DeviceIoControl(pCtx->hDriver, TC_IOCTL_GET_VOLUME_PROPERTIES,
              &tcvp, sizeof(tcvp), &tcvp, sizeof(tcvp), &dwResult, NULL)) {
                continue;
            }
            if (tcvp.fSystemVolume) {
                if ((dwFlags & TCMD_SYSTEM) == 0) {
                    continue;
                }
            } else if (tcvp.fPersistentVolume) {
                if ((dwFlags & TCMD_PERSISTENT) == 0) {
                    continue;
                }
            } else {
                if ((dwFlags & TCMD_DISMOUNTABLE) == 0) {
                    continue;
                }
            }
            dwDrives |= 1<<i;
        }
        *pdwDrives = dwDrives;

    }

    return TRUE;
}

#ifdef _MSC_VER
#  pragma warning( push )
#  pragma warning( disable : 4701 )
#endif
BOOL WINAPI TCSYS_Dismount(PTC_CONTEXT pContext, LPCTSTR pszDrive, DWORD dwFlags)
{
    PTCSYS_CONTEXT       pCtx;
    DWORD                dwResult;
    DWORD                nAttempts;
    TC_DISMOUNT_DATA     tcdd;
    DEV_BROADCAST_VOLUME dbv;
    DEV_BROADCAST_VOLUME dbv_copy;
    BOOL                 fResult;

    TC_DbgPrintf(TEXT("TCSYS_Dismount(\"%s\", 0x%08x)\n"), pszDrive,
      dwFlags);
    TC_Assert(pContext);

    pCtx = (PTCSYS_CONTEXT) pContext->pContext;
    if (pCtx == NULL || pCtx->hDriver == INVALID_HANDLE_VALUE) {
        TC_DbgPutsA("Invalid context or device driver handle!\n");
        return FALSE;
    }

    fResult = TRUE;

    if (dwFlags & TC_WIPE_CACHE) {
        if (!DeviceIoControl(pCtx->hDriver, TC_IOCTL_WIPE_CACHE, NULL, 0,
          NULL, 0, &dwResult, NULL)) {
            TC_DbgPutsA("TC_IOCTL_WIPE_CACHE failed!\n");
            fResult = FALSE;
        }
    }

    dbv.dbcv_size       = sizeof(dbv);
    dbv.dbcv_devicetype = DBT_DEVTYP_VOLUME;
    dbv.dbcv_reserved   = 0;
    dbv.dbcv_flags      = 0;

    if (pszDrive == NULL || *pszDrive == TEXT('\0')) {

        /* Dismount all volumes */

        TC_MOUNT_LIST_DATA tcmld;

        TC_ZeroMemory(&tcmld, sizeof(tcmld));
        if (DeviceIoControl(pCtx->hDriver, TC_IOCTL_GET_MOUNT_LIST, &tcmld,
          sizeof(tcmld), &tcmld, sizeof(tcmld), &dwResult, NULL)) {
            dbv.dbcv_unitmask = tcmld.dwfMountedDrives;
        } else {
            dbv.dbcv_unitmask = 0;
            TC_DbgPutsA("First TC_IOCTL_GET_MOUNT_LIST failed!\n");
            fResult = FALSE;
        }

        for (nAttempts = TC_DISMOUNT_ATTEMPTS; nAttempts; nAttempts--) {
            TC_ZeroMemory(&tcdd, sizeof(tcdd));
            tcdd.nDosDriveNo      = 0;
            tcdd.fIgnoreOpenFiles = (dwFlags & TC_DISMOUNT_FORCIBLY) ? TRUE :
              FALSE;
            dbv_copy = dbv;
            SendMessageTimeout(HWND_BROADCAST, WM_DEVICECHANGE,
              DBT_DEVICEREMOVEPENDING, (LPARAM)&dbv_copy, 0, 1000,
              (PDWORD_PTR)&dwResult);
            if (DeviceIoControl(pCtx->hDriver, TC_IOCTL_DISMOUNT_ALL, &tcdd,
              sizeof(tcdd), &tcdd, sizeof(tcdd), &dwResult, NULL)) {
                if (tcdd.iResult == TC_ERR_FILES_OPEN) {
                    Sleep(TC_DISMOUNT_RETRY_DELAY);
                } else {
                    break;
                }
            }
        }
        if (nAttempts == 0 || tcdd.iResult) {
            TC_DbgPrintfA(
              "TC_IOCTL_DISMOUNT_ALL failed (nAttempts=%d, iResult=0x%08x)\n",
              nAttempts, tcdd.iResult);
            return FALSE;
        }

        TC_ZeroMemory(&tcmld, sizeof(tcmld));
        if (DeviceIoControl(pCtx->hDriver, TC_IOCTL_GET_MOUNT_LIST, &tcmld,
          sizeof(tcmld), &tcmld, sizeof(tcmld), &dwResult, NULL)) {
            dbv.dbcv_unitmask  &= ~tcmld.dwfMountedDrives;
        } else {
            dbv.dbcv_unitmask = 0;
            TC_DbgPutsA("Second TC_IOCTL_GET_MOUNT_LIST failed!\n");
            fResult = FALSE;
        }

    } else {

        /* Dismount single volume */

        INT nDosDriveNo;

        if (!TC_DriveNameIsValid(pszDrive)) return FALSE;
        nDosDriveNo = _totupper(*pszDrive) - TEXT('A');

        dbv.dbcv_unitmask = 1 << nDosDriveNo;

        for (nAttempts = TC_DISMOUNT_ATTEMPTS; nAttempts; nAttempts--) {
            TC_ZeroMemory(&tcdd, sizeof(tcdd));
            tcdd.nDosDriveNo      = nDosDriveNo;
            tcdd.fIgnoreOpenFiles = (dwFlags & TC_DISMOUNT_FORCIBLY) ? TRUE :
              FALSE;
            dbv_copy = dbv;
            SendMessageTimeout(HWND_BROADCAST, WM_DEVICECHANGE,
              DBT_DEVICEREMOVEPENDING, (LPARAM)&dbv_copy, 0, 1000,
              (PDWORD_PTR)&dwResult);
            if (DeviceIoControl(pCtx->hDriver, TC_IOCTL_DISMOUNT, &tcdd,
              sizeof(tcdd), &tcdd, sizeof(tcdd), &dwResult, NULL)) {
                if (tcdd.iResult == TC_ERR_FILES_OPEN) {
                    Sleep(TC_DISMOUNT_RETRY_DELAY);
                } else {
                    break;
                }
            }
        }
        if (nAttempts == 0 || tcdd.iResult) {
            TC_DbgPrintfA(
              "TC_IOCTL_DISMOUNT failed (nAttempts=%d, iResult=0x%08x)\n",
              nAttempts, tcdd.iResult);
            return FALSE;
        }
    }

    if (dbv.dbcv_unitmask) {

        CHAR aszRootDirectory[4];
        INT  i;

        aszRootDirectory[1] = ':';
        aszRootDirectory[2] = '\\';
        aszRootDirectory[3] = '\0';

        for (i = 0; i < 26; i++) {
            if (dbv.dbcv_unitmask & (1 << i)) {
                if (pCtx->fCreateLocalDosDevice) {
                    TCHAR szDeviceName[3]  = TEXT("A:");
                    TCHAR szTargetPath[25] = TEXT("\\Device\\TrueCryptVolumeA");
                    szDeviceName[0]  = TEXT('A') + (TCHAR) i;
                    szTargetPath[23] = *szDeviceName;
                    if (!DefineDosDevice(DDD_RAW_TARGET_PATH |
                      DDD_REMOVE_DEFINITION | DDD_EXACT_MATCH_ON_REMOVE,
                      szDeviceName, szTargetPath)) {
                        fResult = FALSE;
                    }
                }
                aszRootDirectory[0] = 'A' + (CHAR) i;
                SHChangeNotify(SHCNE_DRIVEREMOVED, SHCNF_PATH,
                  aszRootDirectory, NULL);
            }
        }

        SendMessageTimeout(HWND_BROADCAST, WM_DEVICECHANGE,
          DBT_DEVICEREMOVECOMPLETE, (LPARAM)&dbv, 0, 1000,
          (PDWORD_PTR)&dwResult);
    }

    return fResult;
}
#ifdef _MSC_VER
#  pragma warning( pop )
#endif

VOID WINAPI TCSYS_RefreshDriveLetters(PTC_CONTEXT pContext)
{
    PTCSYS_CONTEXT pCtx;
    DWORD          dwMountedDrives;
    DWORD          i;
    TCHAR szDeviceName[3]  = TEXT("A:");
    TCHAR szTargetPath[25] = TEXT("\\Device\\TrueCryptVolumeA");

    TC_Assert(pContext);

    pCtx = (PTCSYS_CONTEXT) pContext->pContext;
    if (pCtx == NULL || pCtx->hDriver == INVALID_HANDLE_VALUE) {
        TC_DbgPrintfA("TCSYS_RefreshDriveLetters()\n");
        TC_DbgPutsA("Invalid context or device driver handle!\n");
        return;
    }

    if (!pCtx->fCreateLocalDosDevice) {
        return;
    }

    TC_DbgPrintfA("TCSYS_RefreshDriveLetters\n");

    if (!TCSYS_GetMountedDrives(pContext, &dwMountedDrives, TCMD_ALL)) {
        TC_DbgPutsA("Cannot enumerate mounted TrueCrypt volumes!\n");
        return;
    }

    for (i = 0; i < 26; i++) {
        szDeviceName[0]  = TEXT('A') + (TCHAR) i;
        szTargetPath[23] = *szDeviceName;
        if (dwMountedDrives & (1 << i)) {
            /* Create symbolic link */
            TCHAR szTemp[MAX_PATH];
            if (!QueryDosDevice(szDeviceName, szTemp, countof(szTemp)) &&
              GetLastError() != ERROR_MORE_DATA) {
                DefineDosDevice(DDD_RAW_TARGET_PATH, szDeviceName, szTargetPath);
            }
        } else {
            /* Remove symbolic link */
            DefineDosDevice(DDD_RAW_TARGET_PATH |
              DDD_REMOVE_DEFINITION | DDD_EXACT_MATCH_ON_REMOVE,
              szDeviceName, szTargetPath);
        }
    }
}

BOOL WINAPI TCSYS_GetVolumeFileName(PTC_CONTEXT pContext, LPCTSTR pszDrive,
  LPTSTR pszDest, DWORD nSize)
{
    UNREFERENCED_PARAMETER(pContext);
    UNREFERENCED_PARAMETER(pszDrive);
    UNREFERENCED_PARAMETER(pszDest);
    UNREFERENCED_PARAMETER(nSize);

    return FALSE;
}

#else /* NO_TCSYS_MOUNT */

BOOL WINAPI TCSYS_Mount(PTC_CONTEXT pContext, LPCTSTR pszFileName,
  LPCSTR paszPassword, LPCTSTR pszDrive, DWORD dwFlags,
  LPCTSTR pszKeyFiles, BOOL fProtectHiddenVolume,
  LPCSTR paszHiddenVolumePassword, LPCTSTR pszHiddenVolumeKeyFiles)
{
    UNREFERENCED_PARAMETER(pContext);
    UNREFERENCED_PARAMETER(pszFileName);
    UNREFERENCED_PARAMETER(paszPassword);
    UNREFERENCED_PARAMETER(pszDrive);
    UNREFERENCED_PARAMETER(dwFlags);
    UNREFERENCED_PARAMETER(pszKeyFiles);
    UNREFERENCED_PARAMETER(fProtectHiddenVolume);
    UNREFERENCED_PARAMETER(paszHiddenVolumePassword);
    UNREFERENCED_PARAMETER(pszHiddenVolumeKeyFiles);

    return FALSE;
}

BOOL WINAPI TCSYS_WipeCache(PTC_CONTEXT pContext)
{
    UNREFERENCED_PARAMETER(pContext);

    return FALSE;
}

BOOL WINAPI TCSYS_GetMountedDrives(PTC_CONTEXT pContext, PDWORD pdwDrives, DWORD dwFlags)
{
    UNREFERENCED_PARAMETER(pContext);
    UNREFERENCED_PARAMETER(pdwDrives);
    UNREFERENCED_PARAMETER(dwFlags);

    return FALSE;
}

VOID WINAPI TCSYS_RefreshDriveLetters(PTC_CONTEXT pContext)
{
    UNREFERENCED_PARAMETER(pContext);
}

BOOL WINAPI TCSYS_Dismount(PTC_CONTEXT pContext, LPCTSTR pszDrive, DWORD dwFlags)
{
    UNREFERENCED_PARAMETER(pContext);
    UNREFERENCED_PARAMETER(pszDrive);
    UNREFERENCED_PARAMETER(dwFlags);

    return FALSE;
}

BOOL WINAPI TCSYS_GetVolumeFileName(PTC_CONTEXT pContext, LPCTSTR pszDrive,
  LPTSTR pszDest, DWORD nSize)
{
    PTCSYS_CONTEXT     pCtx;
    TC_MOUNT_LIST_DATA tcmld;
    DWORD              dwResult;
    int                nDriveNo;
    TCHAR              szFileName[MAX_PATH];

    TC_DbgPrintf(TEXT("TCSYS_GetVolumeFileName(\"%s\")\n"), pszDrive);
    TC_Assert(pContext);

    pCtx = (PTCSYS_CONTEXT) pContext->pContext;
    if (pCtx == NULL || pCtx->hDriver == INVALID_HANDLE_VALUE) {
        TC_DbgPutsA("Invalid context or device driver handle!\n");
        return FALSE;
    }

    if (!TC_DriveNameIsValid(pszDrive)) return FALSE;
    nDriveNo = _totupper(*pszDrive) - TEXT('A');

    TC_ZeroMemory(&tcmld, sizeof(tcmld));
    if (!DeviceIoControl(pCtx->hDriver, TC_IOCTL_GET_MOUNT_LIST, &tcmld,
      sizeof(tcmld), &tcmld, sizeof(tcmld), &dwResult, NULL)) {
        TC_DbgPutsA("TC_IOCTL_GET_MOUNT_LIST failed!\n");
        return FALSE;
    }
    if (!TC_CopyStringFromUnicode(szFileName, MAX_PATH, tcmld.wszVolume[nDriveNo])) {
        return FALSE;
    }

    if (_tcsncmp(szFileName, TEXT("\\??\\"), 4) == 0) {
        return TC_CopyString(pszDest, nSize, szFileName + 4);
    }

    return TC_CopyString(pszDest, nSize, szFileName);
}

#endif /* NO_TCSYS_MOUNT */

#endif /* IMPLEMENT_TC_SYS_INTERFACE */
